"""
Driver for running model on design of experiments cases using Salib sampling methods
"""
import numpy as np

from openmdao.api import DOEDriver, OptionsDictionary
from openmdao.drivers.doe_generators import DOEGenerator

SALIB_NOT_INSTALLED = False
try:
    from SALib.sample import morris as ms
    from SALib.sample import saltelli
except ImportError:
    SALIB_NOT_INSTALLED = True


class SalibDOEGenerator(DOEGenerator):
    def __init__(self):
        if SALIB_NOT_INSTALLED:
            raise RuntimeError(
                "SALib library is not installed. \
                                cf. https://salib.readthedocs.io/en/latest/getting-started.html"
            )
        self._cases = np.array([])
        self._pb = None
        self.called = False

    def __call__(self, design_vars, model=None):
        bounds = []
        names = []
        for name, meta in design_vars.items():
            size = meta["size"]
            meta_low = meta["lower"]
            meta_high = meta["upper"]
            for j in range(size):
                if isinstance(meta_low, np.ndarray):
                    p_low = meta_low[j]
                else:
                    p_low = meta_low

                if isinstance(meta_high, np.ndarray):
                    p_high = meta_high[j]
                else:
                    p_high = meta_high

                display_name = name.split(".")[-1]
                if size > 1:
                    display_name += str(j)
                names.append(display_name)
                bounds.append((p_low, p_high))

        self._pb = {
            "num_vars": len(names),
            "names": names,
            "bounds": bounds,
            "groups": None,
        }
        self._compute_cases()
        self.called = True
        sample = []
        for i in range(self._cases.shape[0]):
            j = 0
            for name, meta in design_vars.items():
                size = meta["size"]
                sample.append((name, self._cases[i, j : j + size]))
                j += size
            yield sample

    def _compute_cases(self):
        raise RuntimeError("Have to be implemented in subclass.")

    def get_cases(self):
        if not self.called:
            raise RuntimeError("Have to run the driver before getting cases")
        return self._cases

    def get_salib_problem(self):
        if not self.called:
            raise RuntimeError(
                "Have to run the driver before getting the SALib problem"
            )
        return self._pb


class SalibMorrisDOEGenerator(SalibDOEGenerator):
    def __init__(self, n_trajs=2, n_levels=4):
        super(SalibMorrisDOEGenerator, self).__init__()
        # number of trajectories to apply morris method
        self.n_trajs = n_trajs
        # number of grid levels
        self.n_levels = n_levels

    def _compute_cases(self):
        self._cases = ms.sample(self._pb, self.n_trajs, self.n_levels)


class SalibSobolDOEGenerator(SalibDOEGenerator):
    def __init__(self, n_samples=1000, calc_second_order=True):
        super(SalibSobolDOEGenerator, self).__init__()
        # number of samples to generate
        self.n_samples = n_samples
        # whether calculing second order indices
        self.calc_second_order = calc_second_order

    def _compute_cases(self):
        self._cases = saltelli.sample(self._pb, self.n_samples, self.calc_second_order)


class SalibDOEDriver(DOEDriver):
    """
    Baseclass for SALib design-of-experiments Drivers
    """

    def __init__(self, **kwargs):
        super(SalibDOEDriver, self).__init__()

        if SALIB_NOT_INSTALLED:
            raise RuntimeError(
                "SALib library is not installed. \
                                cf. https://salib.readthedocs.io/en/latest/getting-started.html"
            )

        self.options.declare(
            "sa_method_name",
            default="Morris",
            values=["Morris", "Sobol"],
            desc="either Morris or Sobol",
        )
        self.options.declare(
            "sa_doe_options",
            types=dict,
            default={},
            desc="options for given SMT sensitivity analysis method",
        )
        self.options.update(kwargs)

        self.sa_settings = OptionsDictionary()
        if self.options["sa_method_name"] == "Morris":
            self.sa_settings.declare(
                "n_trajs",
                types=int,
                default=2,
                desc="number of trajectories to apply morris method",
            )
            self.sa_settings.declare(
                "n_levels", types=int, default=4, desc="number of grid levels"
            )
            self.sa_settings.update(self.options["sa_doe_options"])
            n_trajs = self.sa_settings["n_trajs"]
            n_levels = self.sa_settings["n_levels"]
            self.options["generator"] = SalibMorrisDOEGenerator(n_trajs, n_levels)
        elif self.options["sa_method_name"] == "Sobol":
            self.sa_settings.declare(
                "n_samples",
                types=int,
                default=500,
                desc="number of samples to generate",
            )
            self.sa_settings.declare(
                "calc_second_order",
                types=bool,
                default=True,
                desc="calculate second-order sensitivities ",
            )
            self.sa_settings.update(self.options["sa_doe_options"])
            n_samples = self.sa_settings["n_samples"]
            calc_snd = self.sa_settings["calc_second_order"]
            self.options["generator"] = SalibSobolDOEGenerator(n_samples, calc_snd)
        else:
            raise RuntimeError(
                "Bad sensitivity analysis method name '{}'".format(
                    self.options["sa_method_name"]
                )
            )

    def _set_name(self):
        self._name = "SALib_DOE_" + self.options["sa_method_name"]

    def get_cases(self):
        return self.options["generator"].get_cases()

    def get_salib_problem(self):
        return self.options["generator"].get_salib_problem()



"""
Driver for running model on design of experiments cases using egobox LHS method
"""
import numpy as np

from openmdao.api import DOEDriver
from openmdao.drivers.doe_generators import DOEGenerator

EGOBOX_NOT_INSTALLED = False
try:
    import egobox as egx
except ImportError:
    EGOBOX_NOT_INSTALLED = True


class EgoboxDOEGenerator(DOEGenerator):
    def __init__(self, n_cases, **kwargs):
        super(EgoboxDOEGenerator, self).__init__(**kwargs)

        if EGOBOX_NOT_INSTALLED:
            raise RuntimeError(
                "egobox library is not installed. Run `pip install egobox` to do so."
            )

        # number of cases
        self.n_cases = n_cases

    def __call__(self, design_vars, model=None):
        x_specs = []
        for name, meta in design_vars.items():
            infos = model.get_io_metadata(includes=name)
            dvs_int = {}
            for absname in infos:
                if name == infos[absname]["prom_name"] and (
                    infos[absname]["tags"] & {"wop:int"}
                ):
                    dvs_int[name] = egx.XType.INT

            size = meta["size"]
            meta_low = meta["lower"]
            meta_high = meta["upper"]

            print(name)
            for j in range(size):
                if isinstance(meta_low, np.ndarray):
                    p_low = meta_low[j]
                else:
                    p_low = meta_low

                if isinstance(meta_high, np.ndarray):
                    p_high = meta_high[j]
                else:
                    p_high = meta_high

                if name in dvs_int:
                    x_specs.append(egx.XSpec(egx.XType.INT, [p_low, p_high]))
                else:
                    x_specs.append(egx.XSpec(egx.XType.FLOAT, [p_low, p_high]))
        cases = egx.lhs(x_specs, self.n_cases)
        sample = []
        for i in range(cases.shape[0]):
            j = 0
            for name, meta in design_vars.items():
                size = meta["size"]
                sample.append((name, cases[i, j : j + size]))
                j += size
            yield sample


class EgoboxDOEDriver(DOEDriver):
    """
    Baseclass for egobox design-of-experiments LHS driver.
    """

    def __init__(self, **kwargs):
        super().__init__()

        if EGOBOX_NOT_INSTALLED:
            raise RuntimeError(
                "egobox library is not installed. Run `pip install egobox` to do so."
            )

        self.options.declare(
            "n_cases", types=int, default=2, desc="number of sample to generate"
        )
        self.options.update(kwargs)
        n_cases = self.options["n_cases"]
        self.options["generator"] = EgoboxDOEGenerator(n_cases=n_cases)

    def _set_name(self):
        self._name = "egobox_DOE_LHS"



import numpy as np
import traceback

from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.core.analysis_error import AnalysisError

ONERASEGO_NOT_INSTALLED = False
try:
    from segomoe.sego_defs import get_sego_options, ExitStatus
    from segomoe.constraint import Constraint
    from segomoe.sego import Sego
except ImportError:
    ONERASEGO_NOT_INSTALLED = True


def to_list(lst, size):
    if not (isinstance(lst, np.ndarray) or isinstance(lst, list)):
        return [lst] * size
    diff_len = len(lst) - size
    if diff_len > 0:
        return lst[0:size]
    elif diff_len < 0:
        return [lst[0]] * size
    else:
        return lst


class OneraSegoDriver(Driver):
    """
    OpenMDAO driver for ONERA SEGOMOE optimizer
    """

    def __init__(self, **kwargs):
        """
        Initialize the driver with the following option.
        """
        super(OneraSegoDriver, self).__init__(**kwargs)

        if ONERASEGO_NOT_INSTALLED:
            raise RuntimeError("Onera SEGOMOE library is not installed.")

        # What we support
        self.supports["optimization"] = True
        self.supports["inequality_constraints"] = True
        self.supports["equality_constraints"] = True
        self.supports["two_sided_constraints"] = True
        self.supports["linear_constraints"] = True

        # What we don't support
        self.supports["multiple_objectives"] = False
        self.supports["active_set"] = False
        self.supports["integer_design_vars"] = False
        self.supports["simultaneous_derivatives"] = False
        self.supports["total_jac_sparsity"] = False
        self.supports["gradients"] = False
        self.supports._read_only = True

        self.opt_settings = {}

    def _declare_options(self):
        self.options.declare(
            "optimizer",
            default="SEGOMOE",
            values=["SEGOMOE"],
            desc="Name of optimizers to use",
        )

    def _setup_driver(self, problem):
        super(OneraSegoDriver, self)._setup_driver(problem)

        self.comm = None

    def run(self, path_hs="", eq_tol={}, ieq_tol={}):
        """
        Optimize the problem using SEGOMOE.

        Parameters
        ----------
        problem : Problem object
            An optimization problem of OpenMDAO framework
        path_hs: string, optional
            path to a directory storing hot_start data
        eq_tol: dict
            Dictionary to define specific tolerance for eq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        ieq_tol: dict
            Dictionary to define specific tolerance for ieq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        """
        model = self._problem().model

        path_hs = ""
        self.eq_tol = eq_tol
        self.ieq_tol = ieq_tol
        self.iter_count = 0
        self.name = f"onera_optimizer_{self.options['optimizer'].lower()}"
        self._sego_vars = []
        self._sego_cons = []
        self._map_con = {}
        self._n_mapped_con = 0

        # Initial Run
        with RecordingDebugging(
            self.options["optimizer"], self.iter_count, self
        ) as rec:
            # Initial Run
            model.run_solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        # Format design variables to suit segomoe implementation
        self._initialize_vars()

        # Format constraints to suit segomoe implementation
        self._initialize_cons()

        # Format option dictionary to suit SEGO implementation
        optim_settings = {}
        for opt, opt_dict in get_sego_options().items():
            optim_settings[opt] = opt_dict["default"]
        n_iter = self.opt_settings["maxiter"]
        optim_settings.update(
            {k: v for k, v in self.opt_settings.items() if k != "maxiter"}
        )

        # In OpenMDAO context, obj and constraints are always evaluated together
        optim_settings["grouped_eval"] = True

        # default model
        mod_obj = {"corr": "squar_exp", "regr": "constant", "normalize": True}

        dim = 0
        for name, meta in self._designvars.items():
            dim += meta["size"]
        print("Designvars dimension: ", dim)
        if dim > 10:
            n_components = 3
            mod_obj["n_components"] = n_components
            mod_obj["name"] = "KPLS" if dim > 20 else "KPLSK"
        else:
            n_components = dim
            mod_obj["name"] = "KRG"
        mod_obj["theta0"] = [1.0] * n_components
        mod_obj["thetaL"] = [0.1] * n_components
        mod_obj["thetaU"] = [10.0] * n_components

        optim_settings["model_type"] = {"obj": mod_obj, "con": mod_obj}

        print(optim_settings)
        # Instanciate a SEGO optimizer
        sego = Sego(
            self._objfunc,
            self._sego_vars,
            const=self._sego_cons,
            optim_settings=optim_settings,
            path_hs=path_hs,
            comm=self.comm,
        )

        # Run the optim
        # exit_flag, x_best, obj_best, dt_opt = sego.run_optim(
        exit_flag, x_best, _, _ = sego.run_optim(n_iter=n_iter)

        # Set optimal parameters
        i = 0
        for name, meta in self._designvars.items():
            size = meta["size"]
            self.set_design_var(name, x_best[0][i : i + size])
            i += size

        with RecordingDebugging(
            self.options["optimizer"], self.iter_count, self
        ) as rec:
            model.run_solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        self.exit_flag = (exit_flag == ExitStatus.valid_sol) or (
            exit_flag == ExitStatus.solution_reached
        )

        return self.exit_flag

    def _initialize_vars(self):
        variables = []
        param_meta = self._designvars
        for name, meta in param_meta.items():
            if meta["size"] > 1:
                if np.isscalar(meta["lower"]):
                    variables += [
                        {
                            "name": name + "_" + str(i),
                            "lb": meta["lower"],
                            "ub": meta["upper"],
                        }
                        for i in range(meta["size"])
                    ]
                else:
                    variables += [
                        {
                            "name": name + "_" + str(i),
                            "lb": meta["lower"][i],
                            "ub": meta["lower"][i],
                        }
                        for i in range(meta["size"])
                    ]
            else:
                variables += [{"name": name, "lb": meta["lower"], "ub": meta["upper"]}]
        self._sego_vars = variables

    def _initialize_cons(self, eq_tol=None, ieq_tol=None):
        """
        Format OpenMDAO constraints to suit SEGOMOE implementation

        Parameters
        ----------
        eq_tol: dict
            Dictionary to define specific tolerance for eq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        ieq_tol: dict
            Dictionary to define specific tolerance for ieq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        """

        # for short
        add_con = self._sego_cons.append

        # Start at 1 -> obj first
        eval_index = 1

        con_meta = self._cons
        eq_cons = {name: con for name, con in con_meta.items() if con["equals"]}
        ieq_cons = {name: con for name, con in con_meta.items() if not con["equals"]}

        # Equality constraints
        for name in eq_cons.keys():
            meta = con_meta[name]
            size = meta["size"]
            equals = to_list(meta["equals"], size)
            tol = to_list(eq_tol[name] if name in self.eq_tol else 1e-4, size)

            for k in range(size):
                add_con(
                    Constraint("=", equals[k], name=(name + "_" + str(k)), tol=tol[k])
                )
            # Map eq constraints - range to allow double-sided constraint support
            self._map_con[name] = range(eval_index, eval_index + size)
            eval_index += size
            self._n_mapped_con += size

        # Inequality constraints
        for name in ieq_cons.keys():
            meta = con_meta[name]
            size = meta["size"]
            # Bounds - double sided is supported
            lower = to_list(meta["lower"], size)
            upper = to_list(meta["upper"], size)
            tol = to_list(ieq_tol[name] if name in self.ieq_tol else 1e-4, size)
            # Index list is updated for each real constraint to handle double sided constraints
            index_list = []

            for k in range(size):
                if lower[k] is None or lower[k] < -1e29:
                    # g(x) < upper
                    add_con(
                        Constraint(
                            "<", upper[k], name=(name + "_" + str(k)), tol=tol[k]
                        )
                    )
                    index_list.append(eval_index)
                    eval_index += 1
                    self._n_mapped_con += 1
                elif upper[k] is None or upper[k] > 1e29:
                    # g(x) > lower
                    add_con(
                        Constraint(
                            ">", lower[k], name=(name + "_" + str(k)), tol=tol[k]
                        )
                    )
                    index_list.append(eval_index)
                    eval_index += 1
                    self._n_mapped_con += 1
                else:
                    # lower < g(x) < upper
                    # Split is currently required for sego
                    add_con(
                        Constraint(
                            "<", upper[k], name=(name + "_" + str(k) + "<"), tol=tol[k]
                        )
                    )
                    add_con(
                        Constraint(
                            ">", lower[k], name=(name + "_" + str(k) + ">"), tol=tol[k]
                        )
                    )
                    index_list.append([eval_index, eval_index + 1])
                    eval_index += 2
                    self._n_mapped_con += 2
            self._map_con[name] = index_list

    def _objfunc(self, point):
        """
        Function that evaluates and returns the objective function and the
        constraints. This function is called by SEGOMOE

        Parameters
        ----------
        point : numpy.ndarray
            point to evaluate

        Returns
        -------
        func_dict : dict
            Dictionary of all functional variables evaluated at design point.
        fail : int
            0 for successful function evaluation
            1 for unsuccessful function evaluation
        """
        fail = False
        res = np.zeros(1 + self._n_mapped_con)
        model = self._problem().model

        try:
            # Pass in new parameters
            i = 0

            for name, meta in self._designvars.items():
                size = meta["size"]
                self.set_design_var(name, point[i : i + size])
                i += size

            # Execute the model
            with RecordingDebugging(
                self.options["optimizer"], self.iter_count, self
            ) as _:
                self.iter_count += 1
                try:
                    model._solve_nonlinear()

                # Let the optimizer try to handle the error
                except AnalysisError:
                    model._clear_iprint()
                    fail = True

            # Get the objective function evaluation - single obj support
            for name, obj in self.get_objective_values().items():
                res[0] = obj

            # Get the constraint evaluations:
            for name, con_res in self.get_constraint_values().items():
                # Make sure con_res is array_like
                con_res = to_list(con_res, len(self._map_con[name]))
                # Perform mapping
                for i, con_index in enumerate(self._map_con[name]):
                    if isinstance(con_index, list):
                        # Double sided inequality constraint -> duplicate response
                        for k in con_index:
                            res[k] = con_res[i]
                    else:
                        res[con_index] = con_res[i]

        except Exception as msg:
            tb = traceback.format_exc()

            # Exceptions seem to be swallowed by the C code, so this
            # should give the user more info than the dreaded "segfault"
            print("Exception: %s" % str(msg))
            print(70 * "=", tb, 70 * "=")
            fail = True

        return res, fail



"""Define the NonlinearBlockGS class."""

import os
import numpy as np

from openmdao.core.analysis_error import AnalysisError
from openmdao.recorders.recording_iteration_stack import Recording
from openmdao.api import NonlinearBlockGS
from openmdao import __version__ as openmdao_version


class RecklessNonlinearBlockGS(NonlinearBlockGS):
    """
    Extends Nonlinear block Gauss-Seidel solver with convergence variables options.
    Those options allows to focus on a subset of variables to drive the convergence.
    It allows to get quickest convergence by ignoring 'noise' coming highly non linear
    variables. Obviously the user has to know what he/she is doing because
    in that case some of the variables may not be converged properly
    (hence the 'reckless' prefix in the name).

    Attributes
    ----------
    _convrg_vars: list of string
        List of absolute variable names used to compute relative error and control
        solver convergence.
    _convrg_rtols: list of float
        List of relative error tolerance values for each variables of _convrg_vars. If not set, rtol
        value is used for all specified variables. Only used if _convrg_vars is set.
    """

    SOLVER = "NL: RNLBGS"

    def __init__(self, **kwargs):
        """
        Initialize all attributes.

        Parameters
        ----------
        **kwargs : dict
            options dictionary.
        """
        super(RecklessNonlinearBlockGS, self).__init__(**kwargs)

        self._convrg_vars = None
        self._convrg_rtols = None

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        super(RecklessNonlinearBlockGS, self)._declare_options()
        self.options.declare(
            "convrg_vars",
            types=list,
            default=[],
            desc="list of variables (names) used by relative error criterium.",
        )
        self.options.declare(
            "convrg_rtols",
            types=list,
            default=[],
            desc="list of relative error tolerances corresponding to each"
            " variable specified in convrg_vars option (rtol is used otherwise)",
        )

    def _solve(self):
        """
        Run the iterative solver.

        Overrides opendmao/solvers/solver.py to implement _is_rtol_converged
        """
        maxiter = self.options["maxiter"]
        atol = self.options["atol"]
        iprint = self.options["iprint"]

        self._mpi_print_header()

        self._iter_count = 0
        self._iter_initialize()

        if self._convrg_vars:
            # get initial value of convrg_vars
            norm, val_convrg_vars_minus1 = self._iter_get_norm()
            ratio = 1
        else:
            norm = self._iter_get_norm()
            norm0 = norm if norm != 0.0 else 1.0
            ratio = norm / norm0

        self._mpi_print(self._iter_count, norm, ratio)
        is_rtol_converged = self._is_rtol_converged(ratio)

        while self._iter_count < maxiter and norm > atol and not is_rtol_converged:
            with Recording(type(self).__name__, self._iter_count, self) as rec:
                self._single_iteration()
                self._iter_count += 1
                self._run_apply()

                if self._convrg_vars:
                    norm, val_convrg_vars = self._iter_get_norm()
                    rec.abs = norm
                    rec.rel = (
                        np.abs(val_convrg_vars - val_convrg_vars_minus1)
                        / val_convrg_vars_minus1
                    )
                    ratio = (
                        np.abs(val_convrg_vars - val_convrg_vars_minus1)
                        / val_convrg_vars_minus1
                    )
                    ratio = np.max(ratio)

                    val_convrg_vars_minus1 = val_convrg_vars
                else:
                    norm = self._iter_get_norm()
                    # With solvers, we want to record the norm AFTER the call, but the call needs to
                    # be wrapped in the with for stack purposes, so we locally assign  norm & norm0
                    # into the class.
                    rec.abs = norm
                    rec.rel = norm / norm0
                    ratio = norm / norm0
                    if norm0 == 0:
                        norm0 = 1

            self._mpi_print(self._iter_count, norm, ratio)
            is_rtol_converged = self._is_rtol_converged(ratio)

        system = self._system()
        if system.comm.rank == 0 or os.environ.get("USE_PROC_FILES"):
            prefix = self._solver_info.prefix + self.SOLVER
            is_rtol_converged = self._is_rtol_converged(ratio)
            # Solver terminated early because a Nan in the norm doesn't satisfy the while-loop
            # conditionals.
            if np.isinf(norm) or np.isnan(norm):
                msg = (
                    "Solver '{}' on system '{}': residuals contain 'inf' or 'NaN' after {} "
                    + "iterations."
                )
                if iprint > -1:
                    print(
                        prefix
                        + msg.format(self.SOLVER, system.pathname, self._iter_count)
                    )

                # Raise AnalysisError if requested.
                if self.options["err_on_non_converge"]:
                    raise AnalysisError(
                        msg.format(self.SOLVER, system.pathname, self._iter_count)
                    )

            # Solver hit maxiter without meeting desired tolerances.
            elif norm > atol and not is_rtol_converged:
                msg = "Solver '{}' on system '{}' failed to converge in {} iterations."

                if iprint > -1:
                    print(
                        prefix
                        + msg.format(self.SOLVER, system.pathname, self._iter_count)
                    )

                # Raise AnalysisError if requested.
                if self.options["err_on_non_converge"]:
                    raise AnalysisError(
                        msg.format(self.SOLVER, system.pathname, self._iter_count)
                    )

            # Solver converged
            elif iprint == 1:
                print(prefix + " Converged in {} iterations".format(self._iter_count))
            elif iprint == 2:
                print(prefix + " Converged")

    def _iter_initialize(self):
        """
        Perform any necessary pre-processing operations.

        Returns
        -------
        float
            initial error.
        float
            error at the first iteration.
        """
        self._convrg_vars = self.options["convrg_vars"]
        if self._convrg_vars and not self.options["convrg_rtols"]:
            rtol = self.options["rtol"]
            self._convrg_rtols = rtol * np.ones(len(self._convrg_vars))
        else:
            self._convrg_rtols = self.options["convrg_rtols"]
            if len(self._convrg_rtols) != len(self._convrg_vars):
                raise RuntimeError(
                    "Convergence rtols bad size : should be {}, found {}.".format(
                        len(self._convrg_vars), len(self._convrg_rtols)
                    )
                )

        return super(RecklessNonlinearBlockGS, self)._iter_initialize()

    def _is_rtol_converged(self, ratio):
        """
        Check convergence regarding relative error tolerance.

        Parameters
        ----------
        norm : float
            error (residuals norm)
        norm0 : float
            initial error

        Returns
        -------
        bool
            whether convergence is reached regarding relative error tolerance
        """
        system = self._system()
        if self._convrg_vars:
            nbvars = len(self._convrg_vars)
            rerrs = np.ones(nbvars)
            outputs = np.ones(nbvars)
            for i, name in enumerate(self._convrg_vars):
                outputs = self._get_views_array(system._outputs._views[name])
                residual = self._get_views_array(system._residuals._views[name])
                rerrs[i] = np.linalg.norm(residual) / np.linalg.norm(outputs)
            is_rtol_converged = (rerrs < self._convrg_rtols).all()
            is_rtol_converged = ratio < self.options["rtol"]
        else:
            is_rtol_converged = ratio < self.options["rtol"]
        return is_rtol_converged

    def _iter_get_norm(self):
        """
        Return the norm of the residual regarding convergence variable settings.

        Returns
        -------
        float
            norm.
        """
        system = self._system()
        if self._convrg_vars:
            total = []
            val_convrg_vars = np.zeros(len(self._convrg_vars))
            for i, name in enumerate(self._convrg_vars):
                total.append(system._residuals._views_flat[name])
                val_convrg_vars[i] = np.linalg.norm(
                    self._get_views_array(system._outputs._views[name])
                )
            norm = np.linalg.norm(np.concatenate(total))
        else:
            norm = super(RecklessNonlinearBlockGS, self)._iter_get_norm()

        if self._convrg_vars:
            return norm, val_convrg_vars
        else:
            return norm

    @staticmethod
    def _get_views_array(vector_views):
        """Simple backward compatibility function as views change
        from np.array to (np.array, bool) in openmdao 3.38"""
        if openmdao_version > "3.37.0":
            return vector_views[0]
        else:
            return vector_views



"""
Driver for running model on design of experiments cases using SMT sampling methods
"""
import numpy as np

from openmdao.api import DOEDriver
from openmdao.drivers.doe_generators import DOEGenerator

SMT_NOT_INSTALLED = False
try:
    from smt.sampling_methods import FullFactorial, LHS, Random

    _SAMPLING_METHODS = {"FullFactorial": FullFactorial, "LHS": LHS, "Random": Random}

except ImportError:
    SMT_NOT_INSTALLED = True


class SmtDOEGenerator(DOEGenerator):
    def __init__(self, sampling_method_name, n_cases, **kwargs):
        super(SmtDOEGenerator, self).__init__()

        if SMT_NOT_INSTALLED:
            raise RuntimeError(
                "SMT library is not installed. Run `pip install smt`. See https://smt.readthedocs.io/en/latest"
            )

        # number of trajectories to apply morris method
        self.sampling_method_name = sampling_method_name
        # number of grid levels
        self.n_cases = n_cases
        # options
        self.sampling_method_opts = kwargs

    def __call__(self, design_vars, model=None):
        xlimits = []
        for name, meta in design_vars.items():
            size = meta["size"]
            meta_low = meta["lower"]
            meta_high = meta["upper"]
            for j in range(size):
                if isinstance(meta_low, np.ndarray):
                    p_low = meta_low[j]
                else:
                    p_low = meta_low

                if isinstance(meta_high, np.ndarray):
                    p_high = meta_high[j]
                else:
                    p_high = meta_high

                xlimits.append((p_low, p_high))

        sampling = _SAMPLING_METHODS[self.sampling_method_name](
            xlimits=np.array(xlimits), **(self.sampling_method_opts)
        )
        cases = sampling(self.n_cases)
        sample = []
        for i in range(cases.shape[0]):
            j = 0
            for name, meta in design_vars.items():
                size = meta["size"]
                sample.append((name, cases[i, j : j + size]))
                j += size
            yield sample


class SmtDOEDriver(DOEDriver):
    """
    Baseclass for SMT design-of-experiments Drivers.
    """

    def __init__(self, **kwargs):
        super(SmtDOEDriver, self).__init__()

        if SMT_NOT_INSTALLED:
            raise RuntimeError(
                "SMT library is not installed. cf. https://https://smt.readthedocs.io/en/latest"
            )

        self.options.declare(
            "sampling_method_name",
            types=str,
            default="LHS",
            desc="either LHS, FullFactorial or Ramdom",
        )
        self.options.declare(
            "n_cases", types=int, default=2, desc="number of sample to generate"
        )
        self.options.declare(
            "sampling_method_options",
            types=dict,
            default={},
            desc="options for given SMT sampling method",
        )

        self.options.update(kwargs)
        name = self.options["sampling_method_name"]
        n_cases = self.options["n_cases"]
        opts = self.options["sampling_method_options"]
        self.options["generator"] = SmtDOEGenerator(
            sampling_method_name=name, n_cases=n_cases, **opts
        )

    def _set_name(self):
        self._name = "SMT_DOE_" + self.options["sampling_method_name"]
